@@ -1,7 +1,7 @@
 # frozen_string_literal: true
 
 # Redmine - project management software
-# Copyright (C) 2006-2022  Jean-Philippe Lang
+# Copyright (C) 2006-  Jean-Philippe Lang
 #
 # This program is free software; you can redistribute it and/or
 # modify it under the terms of the GNU General Public License
@@ -30,7 +30,7 @@ class QueryColumn
     self.totalable = options[:totalable] || false
     self.default_order = options[:default_order]
     @inline = options.key?(:inline) ? options[:inline] : true
-    @caption_key = options[:caption] || "field_#{name}".to_sym
+    @caption_key = options[:caption] || :"field_#{name}"
     @frozen = options[:frozen]
   end
 
@@ -104,11 +104,19 @@ class TimestampQueryColumn < QueryColumn
   end
 end
 
+class WatcherQueryColumn < QueryColumn
+  def value_object(object)
+    return nil unless User.current.allowed_to?(:"view_#{object.class.name.underscore}_watchers", object.try(:project))
+
+    super
+  end
+end
+
 class QueryAssociationColumn < QueryColumn
   def initialize(association, attribute, options={})
     @association = association
     @attribute = attribute
-    name_with_assoc = "#{association}.#{attribute}".to_sym
+    name_with_assoc = :"#{association}.#{attribute}"
     super(name_with_assoc, options)
   end
 
@@ -126,10 +134,13 @@ end
 
 class QueryCustomFieldColumn < QueryColumn
   def initialize(custom_field, options={})
-    self.name = "cf_#{custom_field.id}".to_sym
-    self.sortable = custom_field.order_statement || false
-    self.totalable = options.key?(:totalable) ? !!options[:totalable] : custom_field.totalable?
-    @inline = custom_field.full_width_layout? ? false : true
+    name = :"cf_#{custom_field.id}"
+    super(
+      name,
+      :sortable => custom_field.order_statement || false,
+      :totalable =>  options.key?(:totalable) ? !!options[:totalable] : custom_field.totalable?,
+      :inline => custom_field.full_width_layout? ? false : true
+    )
     @cf = custom_field
   end
 
@@ -150,7 +161,8 @@ class QueryCustomFieldColumn < QueryColumn
   end
 
   def value_object(object)
-    if custom_field.visible_by?(object.project, User.current)
+    project = object.project if object.respond_to?(:project)
+    if custom_field.visible_by?(project, User.current)
       cv = object.custom_values.select {|v| v.custom_field_id == @cf.id}
       cv.size > 1 ? cv.sort_by {|e| e.value.to_s} : cv.first
     else
@@ -177,7 +189,7 @@ end
 class QueryAssociationCustomFieldColumn < QueryCustomFieldColumn
   def initialize(association, custom_field, options={})
     super(custom_field, options)
-    self.name = "#{association}.cf_#{custom_field.id}".to_sym
+    self.name = :"#{association}.cf_#{custom_field.id}"
     # TODO: support sorting by association custom field
     self.sortable = false
     self.groupable = false
@@ -207,7 +219,7 @@ class QueryFilter
   def initialize(field, options)
     @field = field.to_s
     @options = options
-    @options[:name] ||= l(options[:label] || "field_#{field}".gsub(/_id$/, ''))
+    @options[:name] ||= l(options[:label] || "field_#{field}".delete_suffix('_id'))
     # Consider filters with a Proc for values as remote by default
     @remote = options.key?(:remote) ? options[:remote] : options[:values].is_a?(Proc)
   end
@@ -235,7 +247,7 @@ class QueryFilter
   end
 end
 
-class Query < ActiveRecord::Base
+class Query < ApplicationRecord
   class StatementInvalid < ::ActiveRecord::StatementInvalid
   end
 
@@ -253,11 +265,12 @@ class Query < ActiveRecord::Base
   has_and_belongs_to_many :roles, :join_table => "#{table_name_prefix}queries_roles#{table_name_suffix}", :foreign_key => "query_id"
   serialize :filters
   serialize :column_names
-  serialize :sort_criteria, Array
-  serialize :options, Hash
+  serialize :sort_criteria, type: Array
+  serialize :options, type: Hash
 
   validates_presence_of :name
   validates_length_of :name, :maximum => 255
+  validates_length_of :description, :maximum => 255
   validates :visibility, :inclusion => {:in => [VISIBILITY_PUBLIC, VISIBILITY_ROLES, VISIBILITY_PRIVATE]}
   validate :validate_query_filters
   validate do |query|
@@ -302,6 +315,7 @@ class Query < ActiveRecord::Base
     "t-"  => :label_ago,
     "~"   => :label_contains,
     "!~"  => :label_not_contains,
+    "*~"  => :label_contains_any_of,
     "^"   => :label_starts_with,
     "$"   => :label_ends_with,
     "=p"  => :label_any_issues_in_project,
@@ -309,18 +323,24 @@ class Query < ActiveRecord::Base
     "!p"  => :label_no_issues_in_project,
     "*o"  => :label_any_open_issues,
     "!o"  => :label_no_open_issues,
+    "ev"  => :label_has_been,       # "ev" stands for "ever"
+    "!ev" => :label_has_never_been,
+    "cf"  => :label_changed_from
   }
 
   class_attribute :operators_by_filter_type
   self.operators_by_filter_type = {
     :list => [ "=", "!" ],
-    :list_status => [ "o", "=", "!", "c", "*" ],
+    :list_with_history =>  [ "=", "!", "ev", "!ev", "cf" ],
+    :list_status => [ "o", "=", "!", "ev", "!ev", "cf", "c", "*" ],
     :list_optional => [ "=", "!", "!*", "*" ],
+    :list_optional_with_history => [ "=", "!", "ev", "!ev", "cf", "!*", "*" ],
     :list_subprojects => [ "*", "!*", "=", "!" ],
     :date => [ "=", ">=", "<=", "><", "<t+", ">t+", "><t+", "t+", "nd", "t", "ld", "nw", "w", "lw", "l2w", "nm", "m", "lm", "y", ">t-", "<t-", "><t-", "t-", "!*", "*" ],
     :date_past => [ "=", ">=", "<=", "><", ">t-", "<t-", "><t-", "t-", "t", "ld", "w", "lw", "l2w", "m", "lm", "y", "!*", "*" ],
-    :string => [ "~", "=", "!~", "!", "^", "$", "!*", "*" ],
-    :text => [  "~", "!~", "^", "$", "!*", "*" ],
+    :string => [ "~", "*~", "=", "!~", "!", "^", "$", "!*", "*" ],
+    :text => [  "~", "*~", "!~", "^", "$", "!*", "*" ],
+    :search => [ "~", "*~", "!~" ],
     :integer => [ "=", ">=", "<=", "><", "!*", "*" ],
     :float => [ "=", ">=", "<=", "><", "!*", "*" ],
     :relation => ["=", "!", "=p", "=!p", "!p", "*o", "!o", "!*", "*"],
@@ -335,6 +355,8 @@ class Query < ActiveRecord::Base
   # Permission required to view the queries, set on subclasses.
   class_attribute :view_permission
 
+  class_attribute :layout, default: 'base'
+
   # Scope of queries that are global or on the given project
   scope :global_or_on_project, (lambda do |project|
     where(:project_id => (project.nil? ? nil : [nil, project.id]))
@@ -368,16 +390,17 @@ class Query < ActiveRecord::Base
       scope.where("#{table_name}.visibility <> ? OR #{table_name}.user_id = ?", VISIBILITY_PRIVATE, user.id)
     elsif user.memberships.any?
       scope.where(
-        "#{table_name}.visibility = ?" +
-          " OR (#{table_name}.visibility = ? AND #{table_name}.id IN (" +
-          "SELECT DISTINCT q.id FROM #{table_name} q" +
-          " INNER JOIN #{table_name_prefix}queries_roles#{table_name_suffix} qr on qr.query_id = q.id" +
-          " INNER JOIN #{MemberRole.table_name} mr ON mr.role_id = qr.role_id" +
-          " INNER JOIN #{Member.table_name} m ON m.id = mr.member_id AND m.user_id = ?" +
-          " INNER JOIN #{Project.table_name} p ON p.id = m.project_id AND p.status <> ?" +
-          " WHERE q.project_id IS NULL OR q.project_id = m.project_id))" +
-          " OR #{table_name}.user_id = ?",
-        VISIBILITY_PUBLIC, VISIBILITY_ROLES, user.id, Project::STATUS_ARCHIVED, user.id)
+        "#{table_name}.visibility = ?" \
+        " OR (#{table_name}.visibility = ? AND EXISTS (SELECT 1" \
+        " FROM #{table_name_prefix}queries_roles#{table_name_suffix} qr" \
+        " INNER JOIN #{MemberRole.table_name} mr ON mr.role_id = qr.role_id" \
+        " INNER JOIN #{Member.table_name} m ON m.id = mr.member_id AND m.user_id = ?" \
+        " INNER JOIN #{Project.table_name} p ON p.id = m.project_id AND p.status <> ?" \
+        " WHERE qr.query_id = #{table_name}.id" \
+        " AND (#{table_name}.project_id IS NULL OR #{table_name}.project_id = m.project_id)))" \
+        " OR #{table_name}.user_id = ?",
+        VISIBILITY_PUBLIC, VISIBILITY_ROLES, user.id, Project::STATUS_ARCHIVED, user.id
+      )
     elsif user.logged?
       scope.where("#{table_name}.visibility = ? OR #{table_name}.user_id = ?", VISIBILITY_PUBLIC, user.id)
     else
@@ -396,7 +419,7 @@ class Query < ActiveRecord::Base
       true
     when VISIBILITY_ROLES
       if project
-        (user.roles_for_project(project) & roles).any?
+        user.roles_for_project(project).intersect?(roles)
       else
         user.memberships.joins(:member_roles).where(:member_roles => {:role_id => roles.map(&:id)}).any?
       end
@@ -501,7 +524,7 @@ class Query < ActiveRecord::Base
 
       add_filter_error(field, :blank) unless
           # filter requires one or more values
-          (values_for(field) and !values_for(field).first.blank?) or
+          (values_for(field) and values_for(field).first.present?) or
           # filter doesn't require any value
           ["o", "c", "!*", "*", "nd", "t", "ld", "nw", "w", "lw", "l2w", "nm", "m", "lm", "y", "*o", "!o"].include? operator_for(field)
     end if filters
@@ -541,7 +564,7 @@ class Query < ActiveRecord::Base
       if has_filter?(field) || !filter.remote
         options[:values] = filter.values
         if options[:values] && values_for(field)
-          missing = Array(values_for(field)).select(&:present?) - options[:values].map{|v| v[1]}
+          missing = Array(values_for(field)).select(&:present?) - options[:values].pluck(1)
           if missing.any? && respond_to?(method = "find_#{field}_filter_values")
             options[:values] += send(method, missing)
           end
@@ -578,7 +601,7 @@ class Query < ActiveRecord::Base
   end
 
   def subproject_values
-    project.descendants.visible.collect{|s| [s.name, s.id.to_s]}
+    project.descendants.visible.pluck(:name, :id).map {|name, id| [name, id.to_s]}
   end
 
   def principals
@@ -607,7 +630,7 @@ class Query < ActiveRecord::Base
     author_values = []
     author_values << ["<< #{l(:label_me)} >>", "me"] if User.current.logged?
     author_values +=
-      users.sort_by(&:status).
+      users.sort_by{|p| [p.status, p]}.
         collect{|s| [s.name, s.id.to_s, l("status_#{User::LABEL_BY_STATUS[s.status]}")]}
     author_values << [l(:label_user_anonymous), User.anonymous.id.to_s]
     author_values
@@ -617,7 +640,7 @@ class Query < ActiveRecord::Base
     assigned_to_values = []
     assigned_to_values << ["<< #{l(:label_me)} >>", "me"] if User.current.logged?
     assigned_to_values +=
-      (Setting.issue_group_assignment? ? principals : users).sort_by(&:status).
+      (Setting.issue_group_assignment? ? principals : users).sort_by{|p| [p.status, p]}.
         collect{|s| [s.name, s.id.to_s, l("status_#{User::LABEL_BY_STATUS[s.status]}")]}
     assigned_to_values
   end
@@ -640,14 +663,14 @@ class Query < ActiveRecord::Base
     else
       statuses = IssueStatus.all.sorted
     end
-    statuses.collect{|s| [s.name, s.id.to_s]}
+    statuses.pluck(:name, :id).map {|name, id| [name, id.to_s]}
   end
 
   def watcher_values
     watcher_values = [["<< #{l(:label_me)} >>", "me"]]
     if User.current.allowed_to?(:view_issue_watchers, self.project, global: true)
       watcher_values +=
-        principals.sort_by(&:status).
+        principals.sort_by{|p| [p.status, p]}.
           collect{|s| [s.name, s.id.to_s, l("status_#{User::LABEL_BY_STATUS[s.status]}")]}
     end
     watcher_values
@@ -786,9 +809,9 @@ class Query < ActiveRecord::Base
     return [] if available_columns.empty?
 
     # preserve the column_names order
-    cols = (has_default_columns? ? default_columns_names : column_names).collect do |name|
+    cols = (has_default_columns? ? default_columns_names : column_names).filter_map do |name|
       available_columns.find {|col| col.name == name}
-    end.compact
+    end
     available_columns.select(&:frozen?) | cols
   end
 
@@ -826,7 +849,7 @@ class Query < ActiveRecord::Base
 
   def column_names=(names)
     if names
-      names = names.select {|n| n.is_a?(Symbol) || !n.blank?}
+      names = names.select {|n| n.is_a?(Symbol) || n.present?}
       names = names.collect {|n| n.is_a?(Symbol) ? n : n.to_sym}
       if names.delete(:all_inline)
         names = available_inline_columns.map(&:name) | names
@@ -845,7 +868,7 @@ class Query < ActiveRecord::Base
   end
 
   def has_custom_field_column?
-    columns.any? {|column| column.is_a? QueryCustomFieldColumn}
+    columns.any?(QueryCustomFieldColumn)
   end
 
   def has_default_columns?
@@ -983,9 +1006,9 @@ class Query < ActiveRecord::Base
         end
       end
 
-      if field == 'project_id' || (self.type == 'ProjectQuery' && %w[id parent_id].include?(field))
+      if field == 'project_id' || (is_a?(ProjectQuery) && %w[id parent_id].include?(field))
         if v.delete('mine')
-          v += User.current.memberships.map(&:project_id).map(&:to_s)
+          v += User.current.memberships.pluck(:project_id).map(&:to_s)
         end
         if v.delete('bookmarks')
           v += User.current.bookmarked_project_ids
@@ -1064,7 +1087,7 @@ class Query < ActiveRecord::Base
   end
 
   def display_type=(type)
-    unless type || self.available_display_types.include?(type)
+    unless type && self.available_display_types.include?(type)
       type = self.available_display_types.first
     end
     options[:display_type] = type
@@ -1076,7 +1099,7 @@ class Query < ActiveRecord::Base
 
   private
 
-  def grouped_query(&block)
+  def grouped_query(&)
     r = nil
     if grouped?
       r = yield base_group_scope
@@ -1099,7 +1122,7 @@ class Query < ActiveRecord::Base
       custom_field = column.custom_field
       send :total_for_custom_field, custom_field, scope
     else
-      send "total_for_#{column.name}", scope
+      send :"total_for_#{column.name}", scope
     end
   rescue ::ActiveRecord::StatementInvalid => e
     raise StatementInvalid.new(e.message)
@@ -1115,13 +1138,13 @@ class Query < ActiveRecord::Base
       group(group_by_statement)
   end
 
-  def total_for_custom_field(custom_field, scope, &block)
+  def total_for_custom_field(custom_field, scope, &)
     total = custom_field.format.total_for_scope(custom_field, scope)
     total = map_total(total) {|t| custom_field.format.cast_total_value(custom_field, t)}
     total
   end
 
-  def map_total(total, &block)
+  def map_total(total, &)
     if total.is_a?(Hash)
       total.each_key {|k| total[k] = yield total[k]}
     else
@@ -1186,7 +1209,6 @@ class Query < ActiveRecord::Base
       "  SELECT customized_id FROM #{CustomValue.table_name}" +
       "  WHERE customized_type='#{target_class}' AND custom_field_id=#{chained_custom_field_id}" +
       "  AND #{sql_for_field(field, operator, value, CustomValue.table_name, 'value', true)}))"
-
   end
 
   def sql_for_custom_field_attribute(field, operator, value, custom_field_id, attribute)
@@ -1265,7 +1287,7 @@ class Query < ActiveRecord::Base
       sql += " OR #{db_table}.#{db_field} = ''" if is_custom_filter || [:text, :string].include?(type_for(field))
     when "*"
       sql = "#{db_table}.#{db_field} IS NOT NULL"
-      sql += " AND #{db_table}.#{db_field} <> ''" if is_custom_filter
+      sql += " AND #{db_table}.#{db_field} <> ''" if is_custom_filter || [:text, :string].include?(type_for(field))
     when ">="
       if [:date, :date_past].include?(type_for(field))
         sql = date_clause(db_table, db_field, parse_date(value.first), nil, is_custom_filter)
@@ -1426,10 +1448,37 @@ class Query < ActiveRecord::Base
       sql = sql_contains("#{db_table}.#{db_field}", value.first)
     when "!~"
       sql = sql_contains("#{db_table}.#{db_field}", value.first, :match => false)
+    when "*~"
+      sql = sql_contains("#{db_table}.#{db_field}", value.first, :all_words => false)
     when "^"
       sql = sql_contains("#{db_table}.#{db_field}", value.first, :starts_with => true)
     when "$"
       sql = sql_contains("#{db_table}.#{db_field}", value.first, :ends_with => true)
+    when "ev", "!ev", "cf"
+      # has been,  has never been, changed from
+      if queried_class == Issue && value.present?
+        neg = (operator.start_with?('!') ? 'NOT' : '')
+        subquery =
+          "SELECT 1 FROM #{Journal.table_name}" +
+          " INNER JOIN #{JournalDetail.table_name} ON #{Journal.table_name}.id = #{JournalDetail.table_name}.journal_id" +
+          " WHERE (#{Journal.visible_notes_condition(User.current, :skip_pre_condition => true)}" +
+          " AND #{Journal.table_name}.journalized_type = 'Issue'" +
+          " AND #{Journal.table_name}.journalized_id = #{db_table}.id" +
+          " AND #{JournalDetail.table_name}.property = 'attr'" +
+          " AND #{JournalDetail.table_name}.prop_key = '#{db_field}'" +
+          " AND " +
+          queried_class.send(:sanitize_sql_for_conditions, ["#{JournalDetail.table_name}.old_value IN (?)", value.map(&:to_s)]) +
+          ")"
+        sql_ev =
+          if %w[ev !ev].include?(operator)
+            " OR " + queried_class.send(:sanitize_sql_for_conditions, ["#{db_table}.#{db_field} IN (?)", value.map(&:to_s)])
+          else
+            ''
+          end
+        sql = "#{neg} (EXISTS (#{subquery})#{sql_ev})"
+      else
+        sql = '1=0'
+      end
     else
       raise QueryError, "Unknown query operator #{operator}"
     end
@@ -1438,32 +1487,41 @@ class Query < ActiveRecord::Base
   end
 
   # Returns a SQL LIKE statement with wildcards
+  #
+  # valid options:
+  # * :match - use NOT LIKE if false
+  # * :starts_with - use LIKE 'value%' if true
+  # * :ends_with - use LIKE '%value' if true
+  # * :all_words - use OR instead of AND if false
+  #   (ignored if :starts_with or :ends_with is true)
   def sql_contains(db_field, value, options={})
     options = {} unless options.is_a?(Hash)
     options.symbolize_keys!
-    prefix = suffix = nil
-    prefix = '%' if options[:ends_with]
-    suffix = '%' if options[:starts_with]
-    if prefix || suffix
-      value = queried_class.sanitize_sql_like value
-      queried_class.sanitize_sql_for_conditions(
-        [Redmine::Database.like(db_field, '?', :match => options[:match]), "#{prefix}#{value}#{suffix}"]
-      )
-    else
-      queried_class.sanitize_sql_for_conditions(
-        ::Query.tokenized_like_conditions(db_field, value, **options)
-      )
-    end
+    queried_class.sanitize_sql_for_conditions(
+      ::Query.tokenized_like_conditions(db_field, value, **options)
+    )
   end
 
   # rubocop:disable Lint/IneffectiveAccessModifier
   def self.tokenized_like_conditions(db_field, value, **options)
     tokens = Redmine::Search::Tokenizer.new(value).tokens
     tokens = [value] unless tokens.present?
+
+    if options[:starts_with]
+      prefix, suffix = nil, '%'
+      logical_opr = ' OR '
+    elsif options[:ends_with]
+      prefix, suffix = '%', nil
+      logical_opr = ' OR '
+    else
+      prefix = suffix = '%'
+      logical_opr = options[:all_words] == false ? ' OR ' : ' AND '
+    end
+
     sql, values = tokens.map do |token|
-      [Redmine::Database.like(db_field, '?', options), "%#{sanitize_sql_like token}%"]
+      [Redmine::Database.like(db_field, '?', options), "#{prefix}#{sanitize_sql_like token}#{suffix}"]
     end.transpose
-    [sql.join(" AND "), *values]
+    [sql.join(logical_opr), *values]
   end
   # rubocop:enable Lint/IneffectiveAccessModifier
 
@@ -1584,7 +1642,7 @@ class Query < ActiveRecord::Base
       else
         from = from - 1 # second
       end
-      if self.class.default_timezone == :utc
+      if ActiveRecord.default_timezone == :utc
         from = from.utc
       end
       s << ("#{table}.#{field} > '%s'" % [quoted_time(from, is_custom_filter)])
@@ -1593,7 +1651,7 @@ class Query < ActiveRecord::Base
       if to.is_a?(Date)
         to = date_for_user_time_zone(to.year, to.month, to.day).end_of_day
       end
-      if self.class.default_timezone == :utc
+      if ActiveRecord.default_timezone == :utc
         to = to.utc
       end
       s << ("#{table}.#{field} <= '%s'" % [quoted_time(to, is_custom_filter)])
