@@ -1,7 +1,7 @@
 # frozen_string_literal: true
 
 # Redmine - project management software
-# Copyright (C) 2006-2023  Jean-Philippe Lang
+# Copyright (C) 2006-  Jean-Philippe Lang
 #
 # This program is free software; you can redistribute it and/or
 # modify it under the terms of the GNU General Public License
@@ -17,21 +17,11 @@
 # along with this program; if not, write to the Free Software
 # Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 
-require File.expand_path('../../test_helper', __FILE__)
+require_relative '../test_helper'
 
 class MailerTest < ActiveSupport::TestCase
   include Redmine::I18n
   include Rails::Dom::Testing::Assertions
-  fixtures :projects, :enabled_modules, :issues, :users, :email_addresses, :user_preferences, :members,
-           :member_roles, :roles, :documents, :attachments, :news,
-           :tokens, :journals, :journal_details, :changesets,
-           :trackers, :projects_trackers,
-           :custom_fields, :custom_fields_trackers,
-           :issue_statuses, :enumerations, :messages, :boards, :repositories,
-           :wikis, :wiki_pages, :wiki_contents, :wiki_content_versions,
-           :versions,
-           :comments,
-           :groups_users, :watchers
 
   def setup
     ActionMailer::Base.deliveries.clear
@@ -199,17 +189,26 @@ class MailerTest < ActiveSupport::TestCase
     end
   end
 
-  def test_email_headers
-    with_settings :mail_from => 'Redmine <redmine@example.net>' do
-      issue = Issue.find(1)
-      Mailer.deliver_issue_add(issue)
+  def test_thumbnail_macro_in_email
+    set_tmp_attachments_directory
+    issue = Issue.generate!(:description => '{{thumbnail(image.png)}}')
+    issue.attachments << Attachment.new(:file => mock_file_with_options(:original_filename => 'image.png'), :author => User.find(1))
+    issue.save!
+
+    assert Mailer.deliver_issue_add(issue)
+    assert_select_email do
+      assert_select 'img[alt="image.png"]'
     end
+  end
+
+  def test_email_headers
+    issue = Issue.find(1)
+    Mailer.deliver_issue_add(issue)
     mail = last_email
     assert_equal 'All', mail.header['X-Auto-Response-Suppress'].to_s
     assert_equal 'auto-generated', mail.header['Auto-Submitted'].to_s
-    # List-Id should not include the display name "Redmine"
-    assert_equal '<redmine.example.net>', mail.header['List-Id'].to_s
     assert_equal 'Bug', mail.header['X-Redmine-Issue-Tracker'].to_s
+    assert_equal 'Low', mail.header['X-Redmine-Issue-Priority'].to_s
   end
 
   def test_email_headers_should_include_sender
@@ -314,6 +313,23 @@ class MailerTest < ActiveSupport::TestCase
     end
   end
 
+  def test_list_id_header_should_include_project_identifier
+    with_settings :mail_from => 'Redmine <redmine@example.net>' do
+      content = WikiContent.find(1)
+      Mailer.deliver_wiki_content_added(content)
+      mail = last_email
+      assert_equal '<ecookbook.redmine.example.net>', mail.header['List-Id'].to_s
+    end
+  end
+
+  def test_list_id_header_excludes_project_identifier_for_non_project_emails
+    with_settings :mail_from => 'Redmine <redmine@example.net>' do
+      Mailer.deliver_test_email(User.find(1))
+      mail = last_email
+      assert_equal '<redmine.example.net>', mail.header['List-Id'].to_s
+    end
+  end
+
   def test_should_not_send_email_without_recipient
     news = News.first
     user = news.author
@@ -658,7 +674,7 @@ class MailerTest < ActiveSupport::TestCase
     ActionMailer::Base.deliveries.clear
     with_settings :notified_events => %w(issue_added) do
       cf = IssueCustomField.generate!
-      issue = Issue.generate!
+      issue = Issue.generate!(:parent => Issue.find(1))
       Mailer.deliver_issue_add(issue)
 
       assert_not_equal 0, ActionMailer::Base.deliveries.size
@@ -667,6 +683,7 @@ class MailerTest < ActiveSupport::TestCase
       assert_mail_body_match /^\* Author: /, mail
       assert_mail_body_match /^\* Status: /, mail
       assert_mail_body_match /^\* Priority: /, mail
+      assert_mail_body_match /^\* Parent task: /, mail
 
       assert_mail_body_no_match /^\* Assignee: /, mail
       assert_mail_body_no_match /^\* Category: /, mail
